#! /usr/bin/perl -w
#
# Hardening script for Debian/Ubuntu Server
# Version: 1.0
#
# (c) CSIRT.SK
#
# Description:
# 
# Used for hardening Ubuntu-based systems in 10 steps:
# 1. Proxy settings
# 2. Firewall settings (iptables, psad)
# 3. Automatic security updates
# 4. Block system accounts
# 5. Harden kernel parameters
# 6. Disable ctrl-alt-del restart
# 7. Harden SSH config
# 8. Disable login banners
# 9. Configure clamav, rkhunter, chkrootkit, maldetect, snort if installed
# 10. Configure daily AV checks and updates
# Before modification, backup of each file (*.bak) is made.

################################################
# CONSTANTS
################################################
#my $scriptsdir = "/home/".`whoami|tr -d "\n"`."/scripts";
my $scriptsdir = ""; #set from argument
my $aptfile = "/etc/apt/apt.conf.d/50unattended-upgrades";
#my $aptfile = "50unattended-upgrades";
my $aptPeriodicFile = "/etc/apt/apt.conf.d/10periodic";
my $fwfile = "firewall.sh";
my $psadfile = "/etc/psad/psad.conf";
#my $psadfile = "psad.conf";
my $envfile = "/etc/environment";
#my $envfile = "environment";
my $aptproxyFile = "/etc/apt/apt.conf.d/02proxy";
#my $aptconfFile = "apt.conf";
my $sysctlfile = "/etc/sysctl.conf";
#my $sysctlfile = "sysctl.conf";
my $restartfile = "/etc/init/control-alt-delete.conf";
#my $restartfile = "control-alt-delete.conf";
my $sshfile = "/etc/ssh/sshd_config";
#my $sshfile = "sshd_config";
my $freshclamfile = "/etc/clamav/freshclam.conf";
#my $freshclamfile = "freshclam.conf";
my $chkrootfile = "/etc/chkrootkit.conf";
#my $chkrootfile = "chkrootkit.conf";
my $rkhunterfile = "/etc/rkhunter.conf";
#my $rkhunterfile = "rkhunter.conf";
my $issuefile = "/etc/issue";
my $netissuefile = "/etc/issue.net";
my $clamavLogdir = "/var/log/clamav";
my $chkrootkitLogdir = "/var/log/chkrootkit";
my $rkhunterLogdir = "/var/log/rkhunter";

################################################
# FUNCTIONS
################################################

#replace a line in file by another line
#	$old - regex of line to replace
#	returns 1 if something's been replaced
sub replaceLine{
	my($filename, $old, $new) = @_;
	my $replaced = 0;

	#open (INFILE,  '<',  "$filename") or die "Can't read file '$filename': $!";
        #open (OUT, '>', "$filename.new") or die "Can't write new file: '$filename': $!";
	die "Error: Cannot backup file '$filename': $!\n" unless system("cp '$filename' '$filename.bak'") == 0;
	
	open (INFILE,  '<',  "$filename.bak") or die "Can't read file '$filename.bak': $!";
	open (OUT, '>', "$filename") or die "Can't write new file: '$filename': $!";

	while( <INFILE> ) { # print the lines before the change
    		#if($_ =~ m|//[[:space:]]+.*-updates|){
    		if($_ =~ m|$old|){
			print OUT "$new\n";	#"nasiel som\n";
			$replaced = 1;
		}
		else{
			print OUT $_;
		}
	}
		
	close INFILE;
	close OUT;
	if($replaced){
		return 1;
	}
	else{
		return 0;
	}
}

#replace occurences of key-value pairs from hash in file
#	$delim:  delimiter to be used
#	$keywords: hash with key-value pairs to insert/replace
sub editFile{
	my ($filename, $keywordsPar, $delim) = @_;
	my $out = $filename;
	my %keywords = %$keywordsPar;
	unless(defined($delim)){ $delim = "=";}

	die "Error: Cannot backup file '$filename': $!\n" unless system("cp '$filename' '$filename.bak'") == 0;
	
	open (INFILE,  '<',  "$filename.bak") or die "Can't read file '$filename.bak': $!";
	open (OUT, '>', "$filename") or die "Can't write new file: '$filename': $!";

	# informacny vypis
	#print "Changing file '$filename':\n";	
	#for(keys %keywords) {
    	#	print("$_ = $keywords{$_}\n");
	#} 

	while( $line = <INFILE> ) { # print the lines before the change
       		#print $out $line;
		my $doplnene = 0;
		# skontroluj, ci sa na riadku nachadza keyword
		for(keys %keywords) {
			if($line =~ /$_/){
				my($kw, $value) = split(/$delim/, $line);
				print OUT "$_$delim$keywords{$_}\n";
				delete($keywords{"$_"});	# po nahradeni zmazeme keyword z hashu
				$doplnene = 1;
				last;
			}
		}
		unless($doplnene){ print OUT $line};
		
    	}

	#doplnenie nenajdenych keywordov
	for(keys %keywords) {
                print OUT "$_$delim$keywords{$_}\n";
	
	}

	close INFILE;
	close OUT;
}

#create $filename.bak backup of $filename
#	$filename: file to backup
sub backupFile{
	my $filename = $_[0];
	die "Error: Cannot backup file '$filename': $!\n" unless system("cp '$filename' '$filename.bak'") == 0;
}

#check required packages
#	returns 1 if everything OK
sub checkPackages{
	my $installed = 1;
	if(system("which psad > /dev/null")){
	#	print "please install package 'psad'\n";
		$installed = 0;
	}
	unless (-f $aptfile){
	#	print "please install package 'unattended-upgrades'\n";
		$installed = 0;
	}
	if(system("which clamscan > /dev/null")){
	#	print "please install package 'clamav'\n";
		$installed = 0;
	}
	if(system("which rkhunter > /dev/null")){
	#	print "please install package 'rkhunter'\n";
		$installed = 0;
	}
	if(system("which chkrootkit > /dev/null")){
	#	print "please install package 'chkrootkit'\n";
		$installed = 0;
	}
	if(! -f "/usr/local/maldetect/maldet"){
		$installed = 0;
	}
	return $installed
}

#interactively set system & APT proxy
sub setProxy{
	print "Set proxy now?(y/n) ";
	my $input = <STDIN>;
	chomp($input);
	if($input eq "y"){
		#system proxy
		print "HTTP proxy IP: ";
		my $httpProxyIP = <STDIN>; chomp($httpProxyIP);
		print "HTTP proxy port: ";
		my $httpProxyPort = <STDIN>; chomp($httpProxyPort);
		print "HTTPS proxy IP: ";
		my $httpsProxyIP = <STDIN>; chomp($httpsProxyIP);
		print "HTTPS proxy port: ";
		my $httpsProxyPort = <STDIN>; chomp($httpsProxyPort);

		my %proxyHash=(
			"export http_proxy" => "\"http://$httpProxyIP:$httpProxyPort\"",
			"export https_proxy" => "\"https://$httpsProxyIP:$httpsProxyPort\"",
			"export HTTP_PROXY" => "\"http://$httpProxyIP:$httpProxyPort\"",
			"export HTTPS_PROXY" => "\"https://$httpsProxyIP:$httpsProxyPort\"",
		);
	
		editFile("$envfile", \%proxyHash);
		print "Changed '$envfile'\n";
	}
	
	print "Set APT proxy?(y/n) ";
	$input = <STDIN>;
	chomp($input);
	if($input eq "y"){
		print "APT proxy IP: ";
		my $httpProxyIP = <STDIN>; chomp($httpProxyIP);
		print "APT proxy port: ";
		my $httpProxyPort = <STDIN>; chomp($httpProxyPort);

		if(-f "$aptproxyFile"){
			die "Error: Cannot backup file '$aptproxyFile': $!\n" unless system("cp '$aptproxyFile' '$aptproxyFile.bak'") == 0;
		}
		open(my $fh, '>', $aptproxyFile) or die "Error: Cannot open file '$aptproxyFile': $!\n";

		print $fh "Acquire::http::Proxy \"http://$httpProxyIP:$httpProxyPort\";\n";
		print "Generated '$aptproxyFile'\n";
		close $fh;
	}	

}

#setup IPTables
sub setIptables{
	if ( -f "$scriptsdir/firewall.sh"){
		print("'$scriptsdir/firewall.sh' already exists, rewrite? (y/n) ");
		my $input = <STDIN>;
		chomp($input);
		unless($input eq "y"){
			print("skipped\n");
			return;
		}
	}
	if (! -d $scriptsdir){
		die "Error: Cannot create directory '$scriptsdir': $!\n" unless system("mkdir -p $scriptsdir") == 0;
	}
	die "Error: Cannot copy file '$fwfile' to '$scriptsdir': $!\n" unless system("cp '$fwfile.default' '$scriptsdir/$fwfile'") == 0;
	die "Error: Cannot copy file 'iptables_load' to '/etc/network/if-pre-up.d': $!\n" unless system("cp iptables_load /etc/network/if-pre-up.d") == 0; 
	die "Error: Cannot change permission on '/etc/network/if-pre-up.d': $!\n" unless system("chmod a+x /etc/network/if-pre-up.d/iptables_load") == 0; 
	print "OK\n";
	print "generated '$scriptsdir/$fwfile'\n";
	print "generated '/etc/network/if-pre-up.d/iptables_load'\n";
	print ">> activate by running sudo sh $scriptsdir/$fwfile\n";
}

#setup PSAD 
#return 1 if ok, 0 if psad not installed
sub setPsad{
	if(system("which psad > /dev/null")){
		return 0;
	}

	my %psadHash=(
		"ENABLE_AUTO_IDS" => "Y;",
		"IPTABLES_BLOCK_METHOD" => "Y;",
	);
	editFile("$psadfile", \%psadHash, " ");
	print "updated '$psadfile'\n";
	return 1;
}

#setup security updates
#return 1 if ok, 0 if unattended-upgrades not installed
sub setUpdates{
	if (-f "$aptfile"){
		#enable security updates
		$ret = replaceLine("$aptfile", "//[[:space:]]+.*-security", "\t".'"${distro_id}:${distro_codename}-security";');
		if($ret == "1"){
			print "updated '$aptfile'\n";
		}else{
			print "security updates already enabled\n";
		}
		#enable periodic updates
		if (-f "$aptPeriodicFile"){
			backupFile("$aptPeriodicFile");
			my $aptperiodicContent = "APT::Periodic::Update-Package-Lists \"1\";\nAPT::Periodic::Download-Upgradeable-Packages \"1\";\nAPT::Periodic::AutocleanInterval \"7\";APT::Periodic::Unattended-Upgrade \"1\";";
			print "$aptperiodicContent";
			die "Error: Cannot modify '$aptPeriodicFile': $!\n" unless system("echo "$aptperiodicContent" > '$aptPeriodicFile'") == 0;
			print("OK\n");
			
		}
		
		return 1;
	}
	return 0;
}

#lock system accounts
sub lockSystemAccounts{

}

#harden kernel parameters
sub hardenKernel{
	my %sysctlHash=(
		"net.ipv4.ip_forward" => "0",
		"net.ipv4.conf.default.rp_filter" => "1",
		"net.ipv4.conf.all.rp_filter" => "1",
		"net.ipv4.tcp_syncookies" => "1",
		"net.ipv4.conf.all.accept_redirects" => "0",
		"net.ipv4.conf.all.secure_redirects" => "0",
		"net.ipv4.conf.all.send_redirects" => "0",
		"net.ipv4.conf.all.accept_source_route" => "0",
		"net.ipv4.conf.all.log_martians" => "1",
		"net.ipv4.conf.default.send_redirects" => "0",
		"net.ipv4.conf.default.accept_source_route"=> "0",
		"net.ipv4.conf.default.accept_redirects" => "0",
		"net.ipv4.conf.default.secure_redirects" => "0",
		"net.ipv4.icmp_echo_ignore_broadcasts"=> "1",
		"net.ipv4.icmp_ignore_bogus_error_responses" => "1",
		"net.ipv6.conf.all.disable_ipv6" => "1",
	);

	editFile("$sysctlfile", \%sysctlHash);
	die "Error: Cannot update sysctl: $!\n" unless system("sysctl -p") == 0;
	print("updated '$sysctlfile'\n");
}

#disable ctrl-alt-delete restart
sub disableRestart{
	my $ret;
	if(-f "$restartfile"){
		#old (< 16.04) way using ctrl-alt-del file
		$ret = replaceLine("$restartfile", "exec shutdown -r now", "#".'exec shutdown -r now "Control-Alt-Delete pressed"');
		if($ret == "1"){
			print "OK\n";
			print "updated '$restartfile'\n";
		}else{
			print "Error: cannot disable ctrl-alt-del restart sequence!\n"; exit 1;
		}

	}else{
		#systemctl
		die "Error: Cannot disable ctrl-alt-delete restart: $!\n" unless system("systemctl mask ctrl-alt-delete.target") == 0;
		print "OK\n";
		print "masked ctrl-alt-del.target...restarting systemd...\n";
		die "Error: Cannot restart systemd: $!\n" unless system("systemctl daemon-reload") == 0;
	}
}

#setup SSHD
sub setSSHD{
	my %sshash=(
	        "PermitRootLogin" => "no",
		"X11Forwarding" => "no",
		"AllowTcpForwarding" => "no",
		"AllowGroups" => "sshLogin",
	);

	if (-f $sshfile){
	        editFile("$sshfile", \%sshash, " ");
	        print "updated '$sshfile'\n";
		print "User to be added to sshLogin group: ";
	        my $sshUser = <STDIN>; chomp($sshUser);
		my $ret = `grep sshLogin /etc/group`;
		if($ret eq ""){
	#        	print "does not exists\n";
			die "Error: Cannot create group 'sshLogin': $!\n" unless system("groupadd sshLogin") == 0;
		}
		else{
			print "groupadd sshLogin skipped (already exists)\n";
		}
		if( $ret =~ "$sshUser"){
			print "adding '$sshUser' to sshLogin skipped (user already in group)\n";
			return;
		}
		die "Error: Cannot add '$sshUser' to group 'sshLogin': $!\n" unless system("adduser $sshUser sshLogin") == 0;
		print "   To configure cert-based login follow these steps:\n";
		print "   -- uncomment line 'AuthorizedKeysFile...' and copy public key to that file (create if idoes not exist)\n";
		print "   -- disable password login by setting 'PasswordAuthentication no'\n";
	}
	else{
        	print "skipped (SSH server not installed)\n";
	}
}

#disable system and network login banners
sub disableBanners{

}

#setup Clamav
#return 1 if ok, 0 if not installed
sub setClamav{
	if(system("which clamscan > /dev/null")){
	#	print "please install package 'clamav'\n";
		return 0;
	}
	#setup proxy
	$ret = `grep "http_proxy" "$envfile" | tr -d '"'`;
	unless($ret eq ""){
		my @str = split('//', $ret);
		my ($proxyIP, $proxyPort) = split(/:/, $str[$#str]);
		#print "ip=$proxyIP\n";
		#print "port=$proxyPort\n";
		my %clamHash=(
			"HTTPProxyServer" => "$proxyIP",
			"HTTPProxyPort" => "$proxyPort",
		);
		editFile("$freshclamfile", \%clamHash, " ");
		print "updated '$freshclamfile'\n";
	}
	#change logdir owner
	die "Error: Cannot change owner of '/var/log/clamav' to 'clamav': $!\n" unless system("chown clamav:clamav /var/log/clamav") == 0;

	return 1;
}

#setup RKHunter
sub setRkhunter{
	if(system("which rkhunter > /dev/null")){
	#	print "please install package 'rkhunter'\n";
		return 0;
	}

	my %rkhunterHash=(
		"LOGFILE" => "$rkhunterLogdir/rkhunter.log",
	);

	if (! -d "$rkhunterLogdir"){
		die "Error: Cannot create directory '$rkhunterLogdir': $!\n" unless system("mkdir $rkhunterLogdir") == 0;
	}
	editFile("$rkhunterfile", \%rkhunterHash);
        print "updated '$rkhunterfile'\n";
	#logrotate
	die "Error: Cannot create RKHunter logrotate file: $!\n" unless system("cp rkhunter.default /etc/logrotate.d/rkhunter") == 0;
	print "generated '/etc/logrotate.d/rkhunter'\n";
	#disable former cronjob
	die "Error: Cannot change permissions on '/etc/cron.daily/rkhunter': $!\n" unless system("chmod -x  /etc/cron.daily/rkhunter") == 0;
	return 1;
}

#setup ChkRootkit
sub setChkrootkit{
	if(system("which chkrootkit > /dev/null")){
	#	print "please install package 'chkrootkit'\n";
		return 0;
	}
	#copy logrotate script
	die "Error: Cannot create CHKRootkit logrotate file: $!\n" unless system("cp chkrootkit.default /etc/logrotate.d/chkrootkit") == 0;
        print "generated '/etc/logrotate.d/chkrootkit'\n";
	#disable former cronjob
	die "Error: Cannot change permissions on '/etc/cron.daily/chkrootkit': $!\n" unless system("chmod -x  /etc/cron.daily/chkrootkit") == 0;
	
	return 1;
}

#setup Maldetect
sub setMaldetect{
	if(! -f "/usr/local/maldetect/maldet"){
		return 0;
	}
	die "Error: Cannot create LMDetect logrotate file: $!\n" unless system("cp maldetect.default /etc/logrotate.d/maldetect") == 0;
        print "generated '/etc/logrotate.d/maldetect'\n";

	return 1;
}

#setup Snort
sub setSnort{
	if(! -f "/usr/sbin/snort"){
		return 0;
	}
	die "Error: Cannot create Snort logrotate file: $!\n" unless system("cp snort.default /etc/logrotate.d/snort") == 0;
        print "generated '/etc/logrotate.d/snort'\n";

	return 1;
}

#setup daily AV checks
sub setDailyChecks{

	die "Error: Cannot copy file 'dennaKontrola.sh.default' to '/etc/cron.daily': $!\n" unless system("cp dennaKontrola.sh.default /etc/cron.daily/dennaKontrola") == 0;
	die "Error: Cannot change permissions on '/etc/cron.daily/dennaKontrola': $!\n" unless system("chmod a+x  /etc/cron.daily/dennaKontrola") == 0;
	print "dennaKontrola: OK\n";

	#print "Deactivate /etc/cron.daily/chkrootkit (to remove duplicity of controls): OK \n";
	#print "Deactivate /etc/cron.daily/rkhunter(to remove duplicity of controls): OK \n";
	#Maldetect cron.daily script is fine
	#change owner of copied scripts
	my $owner = `ls -ld . | cut -d' ' -f3`; chomp $owner;
	print "Error: Cannot change owner of scripts: $!\n" unless system("chown -R $owner:$owner $scriptsdir") == 0;

}

##################################################
# CODE
##################################################
if($#ARGV != 0 || $ARGV[0] eq "-h"){
	print "usage: sudo perl hardening.pl <path>\n";
	print "<path>: path to scripts directory where files will be extracted (will be created if not-existent)\n\n";
	print "Edit *.default files to set correct settings before running the script:\n";
	print "-firewall.sh.default: default iptables config\n";
	print "-dennaKontrola.sh.default: will be run daily with Cron\n";
	print "-chkrootkit.default: logrotate config for Chkrootkit\n";
	print "-rkhunter.default: logrotate config for Rkhunter\n\n";
	print "Note: If you see perl error about \"Setting locale failed...\" run these commands:\n";
	print "locale\n";
	print "sudo locale-gen sk_SK.UTF-8 en_US.UTF-8\n";
	print "sudo dpkg-reconfigure locales\n"; 
	exit 0;
}

print "##############################################\n";
print "# Hardening utility for Debian/Ubuntu Server #\n";
print "# (c) CSIRT.SK                               #\n";
print "##############################################\n";

$scriptsdir = "$ARGV[0]";
my $sysinfo = `uname -a`;
print "Your system is: $sysinfo\n";

=just example
my %hash=(
	"sshUsers" => "user",
	"sshGroups" => "group",
);

editFile("test", \%hash);
=cut

# 1. Proxy
print("\n1. Proxy settings\n");
print("===================================\n");
my $systemProxy = `grep -i 'proxy' "$envfile"`;
chomp($systemProxy);
if($systemProxy eq ""){
	print "no proxy detected\n";
}
else{
	print "detected proxy: '$systemProxy'\n";
}
setProxy();

# 2. FW
# 2.1 IPTables
print("\n2.1 Firewall - IPTables\n");
print("===================================\n");
setIptables();

# 2.2 PSAD
print("\n2.2 Adaptive Firewall - PSAD\n");
print("===================================\n");
if(! setPsad()){
	print("skipped (PSAD not installed)\n");
}

# 3. Automatic security updates
print("\n3. Automatic security updates\n");
print("===================================\n");
if(! setUpdates()){
	print("skipped (unattended-upgrades not installed)\n");
}

# 4. Block system accounts login
print("\n4. Blocking system accounts\n");
print("===================================\n");
$ret = `sh login-block.sh`;
chomp($ret);
if($ret >= 0){
	print("OK\n");
	print("blocked $ret accounts\n");
}
else{
	die "Error: Cannot block accounts: $!\n"
} 

# 5. Kernel configuration
print("\n5. Kernel parameters\n");
print("===================================\n");
hardenKernel();

# 6. Disable ctrl-alt-d restart
print("\n6. Disable ctrl-alt-del restart\n");
print("===================================\n");
disableRestart();

# 7. SSH configuration 
print("\n7. SSH configuration\n");
print("===================================\n");
setSSHD();

# 8. Disable login banners
print("\n8. Disable login banners\n");
print("===================================\n");
backupFile("$issuefile");
die "Error: Cannot modify '$issuefile': $!\n" unless system("echo '' > '$issuefile'") == 0;

backupFile("$netissuefile");
die "Error: Cannot modify '$netissuefile': $!\n" unless system("echo '' > '$netissuefile'") == 0;
print("OK\n");

# 9. AV solutions
print("\n9. AV solutions\n");
print("===================================\n");
if(! setClamav()){
	print("clamav skipped (not installed)\n");
}
if(! setRkhunter()){
	print("rkhunter skipped (not installed)\n");
}
if(! setChkrootkit()){
	print("chkrootkit skipped (not installed)\n");
}
if(! setMaldetect()){
	print("maldetect skipped (not installed)\n");
}

# 10. Place scripts to ~/scripts and set up cron
print("\n10. Set up daily checks\n");
print("===================================\n");
setDailyChecks();

print("Hooray! Script finished successfully.\n\n");

# 0. check for recommended packages
if(checkPackages() == 0){
	print "However, one or more recommended packages are not installed.\n";
	print "Recommended packages:\n";
	print "psad unattended-upgrades clamav rkhunter chkrootkit maldetect\n\n";
}

print("Activate firewall by executing $scriptsdir/firewall.sh\n\n");

print("Consider also the following recommendations. Please note that they are system specific and may require some research before being enabled:\n");
# Other recommendations
print("---------------------------------\n");
print("- Turn off unnecessary services\n");
print("- Protecting bootloader with password\n");
print("- Lock /etc/resolv/conf against changes\n"); 
print("- System integrity checks\n");
print("- Disk encryption\n");

exit 0;
